# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.crawler
~~~~~~~~~~~~~~~~~

Crawl the links on from the provided start point.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from threading import Thread
from queue import Queue
import time

from typing import (List, Tuple, Optional)

from deadlinks.link import Link
from deadlinks.status import Status
from deadlinks.index import Index
from deadlinks.settings import Settings
from deadlinks.exceptions import (
    DeadlinksIgnoredURL,
    DeadlinksRedicrectionURL,
)


class Crawler:
    """ Crawler/Spider Application """

    def __init__(self, settings: Settings) -> None:
        r"""
        Accepts settings and assign them to the members
        in order to start crawling.
        """

        self.settings = settings
        self.retry = settings.retry
        self.index = Index()

        self.crawling = False
        self.queue = Queue() # type: Queue

        # Initialization of the Queue and Index
        self._base = settings.base

        # Checking if this is Ignored URL
        is_ignored, message = self.is_ignored(settings.base)
        if is_ignored:
            self._base.status = Status.IGNORED
            self._base.message = message
            error = "Issue with Base URL <{}>: {}"
            raise DeadlinksIgnoredURL(error.format(self._base.url(), message))

        self.add(self._base)

    def start(self) -> None:
        r""" Starts the crawling process """

        if self.crawling:
            return

        self.crawling = True

        if self.settings.threads > 1:
            for idx in range(1, 1 + self.settings.threads):
                thread = Thread(target=self.indexer, args=[idx], daemon=True)
                thread.start()
            self.queue.join()
        else:
            self.indexer()

    def add(self, link: Link) -> None:
        """ Queue URL. """
        self.index.put(link)
        self.queue.put(link)

    def is_ignored(self, url: Link) -> Tuple[bool, Optional[str]]:
        """ Check if url can be ignored """

        # We can have few different cases when URL is ignored.

        # TODO - Site Owner Ask (via meta tag) to ignore this URL
        # https://support.google.com/webmasters/answer/93710?hl=en

        # TODO - Site Owner Ask (via robota.txt) to ignore this URL
        # https://www.robotstxt.org/robotstxt.html

        # We have an external URL and cheking external URL are Off
        if url.is_external(self.settings.base) and not self.settings.external:
            return (True, "External indexation is Off")

        # This is a URL that fits to one of the ignored domains
        if url.match_domains(self.settings.domains):
            return (True, "Matching ignored domain")

        # This is a URL that fits to one of the ignored pathes
        if url.match_pathes(self.settings.pathes):
            return (True, "Matching ignored path")

        # FORBIDDEN: This is a local URL that located outside indexed path.
        if not url.is_external(self.settings.base) and  \
            self.settings.stay_within_path and not self._base.within(url):
            return (True, "URL outside of the allowed path")

        return (False, None)

    def update(self, url: Link) -> None:
        """ Update state or the url by checking it's data."""

        # We assume that URL Link that passed to this method is in status
        # Status.UNDEFINED, therefor we can perform required checks.
        if url in self.index:
            url = self.index[url]

        if url.status != Status.UNDEFINED:
            return

        # Adding URL to index so we can track its state.
        is_ignored, message = self.is_ignored(url)

        if is_ignored:
            url.status = Status.IGNORED
            url.message = str(message)
            return

        try:
            # TODO - rething short calls implementation.
            is_external = url.is_external(self.settings.base)
            if not url.exists(is_external, retries=self.retry):
                url.status = Status.NOT_FOUND
                return
        except DeadlinksRedicrectionURL as _href:
            # ok, so next time we looking for this
            # we will need to make lookup to redirected URL.
            url.status = Status.REDIRECTION
            self.add_and_go(url, str(_href))
            return
        except DeadlinksIgnoredURL:
            # we catching this exception jic, but "it should never happen"
            return

        # we defining status of this url as FOUND
        url.status = Status.FOUND

        for href in url.links:
            self.add_and_go(url, href)

    def add_and_go(self, url: Link, href: str) -> None:
        """ reducing code duplication"""

        # Create link variable of type Link that represent a new
        #    relative to URL link that has href
        link = Link(url.link(href)) # type: Link

        # adding it to queue
        self.add(link)

        # Update link by adding url as link referrer
        link.add_referrer(url.url())

    def ignores(self) -> bool:
        """ return "ignore" state of the crawler """
        ignore_domains = len(self.settings.domains) > 0
        ignore_pathes = len(self.settings.pathes) > 0
        return ignore_domains or ignore_pathes

    @property
    def ignored(self) -> List[Link]:
        """ Return URLs we have ignore to check. """
        return self.index.ignored()

    @property
    def succeed(self) -> List[Link]:
        """ Return URLs we exists. """
        return self.index.succeed()

    @property
    def failed(self) -> List[Link]:
        """ Return URLs failed to exist. """
        return self.index.failed()

    @property
    def redirected(self) -> List[Link]:
        """ Return URLs failed to exist. """
        return self.index.redirected()

    def indexer(self, thread_number: int = 0) -> None:
        """ Runs indexation operation using piped source. """

        while True:
            while not self.queue.empty():
                url = self.queue.get()
                self.update(url)
                self.queue.task_done()
            else:

                if thread_number == 0:
                    break
                time.sleep(thread_number / 10)


if __name__ == "__main__":

    crawler = Crawler(
        Settings(
            "http://localhost:1313/docs/azure/",
            check_external_urls=False,
            ignore_pathes=["issues/new", "edit/master", "commit/"],
            threads=10,
            retry=None,
        ),
    )
    crawler.start()
    print("RESULTS - TOTAL (", len(crawler.index), ")")

    print("All")
    for k, item in enumerate(crawler.index.succeed()):
        print("{0!s:^8}\t{1!s:^8}\t{2}\t{3}".format(k, item.status, item.url(), item.message))



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.help
~~~~~~~~~~~~~~~~~

better help for --help

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from textwrap import dedent

from click import Command
from click import Context as Ctx
from click import HelpFormatter as Formatter


class CLI(Command):
    """ Custom CLI class, for henerating better `help`"""

    EXAMPLES = dedent(
        """\

            // Check links (including external) at http://localhost:8080/ in 10 threads,
            // but not ones from domians play.golang.org or github.com
            deadlinks localhost:8080 -n 10 -e -d play.golang.org -d github.com

            // Limit check only to links in /docs/ and run crawler in 10 threads.
            deadlinks localhost:1313/docs/ -n 10

            // Run Crawler for all local links that belong to 127.0.0.1:1313
            // using http://localhost:1313/docs/ as start URL.
            deadlinks localhost:1313/docs/ -n 10 --full-site-check
        """)

    def format_help(self, ctx: Ctx, formatter: Formatter) -> None:
        """Writes the help into the formatter if it exists."""

        self.format_help_text(ctx, formatter)
        self.format_usage(ctx, formatter)
        self.format_examples(ctx, formatter)
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)

    def format_help_text(self, ctx: Ctx, formatter: Formatter) -> None:
        """Writes the help text to the formatter if it exists."""

        formatter.write_paragraph()
        formatter.write_text(self.help)

    def format_examples(self, ctx: Ctx, formatter: Formatter) -> None: # pylint: disable-msg=R0201
        """ Writes the examples into the formatter. """

        formatter.write_paragraph()
        with formatter.section('Examples'):
            for line in self.EXAMPLES.split("\n"):
                formatter.write("  {}\n".format(line))
            formatter.write('\n')

    def format_usage(self, ctx: Ctx, formatter: Formatter) -> None:
        """ Writes the usage line into the formatter. """
        pieces = self.collect_usage_pieces(ctx)
        pieces = pieces[::-1]

        formatter.write_paragraph()
        with formatter.section('Usage'):
            formatter.write_text("{} {}".format(
                ctx.command_path,
                ' '.join(pieces).rstrip('\n'),
            ))



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.index
~~~~~~~~~~~~~~~

Provides a collection interface

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Dict, List, Iterator, Callable) #pylint: disable-msg=W0611

from deadlinks.link import Link
from deadlinks.status import Status


class Index:
    r""" Links collection """

    def __init__(self) -> None:
        self._index = dict() # type: Dict[str, Link]

    def __len__(self) -> int:
        """ Find out how many links in this index."""
        return len(self._index)

    def __iter__(self) -> Iterator[Link]:
        """ Iterating over a index. """
        return self._index.values().__iter__()

    def __contains__(self, link: Link) -> bool:
        """ Checks links existance in the index. """
        return link.url() in self._index

    def __getitem__(self, link: Link) -> Link:
        """ """
        return self._index[link.url()]

    def put(self, link: Link) -> None:
        """ Puts a link to the index. """
        if link.url() in self._index:
            return
        self._index[link.url()] = link

    def all(self) -> List[Link]:
        """ Return links in the index (but not UNDEFINED). """

        lmbd = lambda x: x.status != Status.UNDEFINED
        return self._filter(lmbd)

    def succeed(self) -> List[Link]:
        """ Filters succeed urls from index. """

        lmbd = lambda x: x.status == Status.FOUND
        return self._filter(lmbd)

    def redirected(self) -> List[Link]:
        """ Filters succeed urls from index. """

        lmbd = lambda x: x.status == Status.REDIRECTION
        return self._filter(lmbd)

    def failed(self) -> List[Link]:
        """ Filters failed urls from index. """

        lmbd = lambda x: x.status == Status.NOT_FOUND
        return self._filter(lmbd)

    def ignored(self) -> List[Link]:
        """ Filters failed urls from index. """

        lmbd = lambda x: x.status == Status.IGNORED
        return self._filter(lmbd)

    def undefined(self) -> List[Link]:
        """ Filters undefined urls from index. """

        lmbd = lambda x: x.status == Status.UNDEFINED
        return self._filter(lmbd)

    def _filter(self, lambda_func: Callable[[Link], bool]) -> List[Link]:
        """ Filters  values according lambda. """

        return list(filter(lambda_func, self._index.values()))



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.report.console
~~~~~~~~~~~~~~~~~~~~~~~~~

Formats some text

TODO
- [ ] info - add output of ignored domains and pathes


:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Dict, Sequence) #pylint: disable-msg=W0611

import click

from deadlinks.crawler import Crawler


class Console:

    params_colors = {
        'succeed': 'green',
        'failed': 'red',
        'ignored': 'yellow',
    }

    def __init__(self, crawler: Crawler, **opts: Dict) -> None:
        self._crawler = crawler
        self._opts = opts

    def is_colored(self) -> bool:
        return not self._opts.get('no_colors', False)

    def info(self) -> str:
        message = "URL=<{}>; External Cheks={}; Threads={}; Retry={}".format(
            self._crawler.settings.base,
            "On" if self._crawler.settings.external else "Off",
            self._crawler.settings.threads,
            self._crawler.settings.retry,
        )

        return message

    def stats(self) -> str:

        message = "Stats: Found {0}; Not Found {1}; Ignored {2}"
        found = click.style(str(len(self._crawler.succeed)), fg='green')
        not_found = click.style(str(len(self._crawler.failed)), fg='red')
        ignored = click.style(str(len(self._crawler.ignored)), fg='yellow')

        return message.format(found, not_found, ignored)

    def _generate(self, key: str) -> str:
        """ generate a report about urls """
        if key not in {'failed', 'ok', 'ignored'}:
            return ""

        param = 'succeed' if key == 'ok' else key
        links = self._crawler.__getattribute__(param)
        if not links:
            return ""

        param_color = click.style(param, fg=self.params_colors[param])

        return '\n'.join(map(lambda x: "[ {} ] {}".format(param_color, x), links))

    def report(self) -> None:
        """ generate report and pass it to the echo callback """

        info = self.info() # type: str
        stat = self.stats() # type: str

        click.echo(info, color=self.is_colored())
        click.echo("=" * len(click.unstyle(info)))
        click.echo(stat, color=self.is_colored())
        click.echo("-" * len(click.unstyle(stat)))

        # show some url report(s)
        show = list(self._opts.get('show', [])) # type: Sequence[str]

        if 'none' in show:
            return

        if 'all' in show:
            show = [
                'ok',
                'failed',
                'ignored',
            ]

        for report in show:
            OUTPUT_REPORT = self._generate(report)

            if len(OUTPUT_REPORT):
                click.echo(OUTPUT_REPORT, color=self.is_colored())



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.main
~~~~~~~~~~~~~~~~~

main (cli interface)

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import Dict, Any

import click

from deadlinks.help import CLI
from deadlinks.settings import Settings
from deadlinks.link import Link
from deadlinks.crawler import Crawler
from deadlinks.exceptions import DeadlinksExeption
from deadlinks.reports import Console
from deadlinks.__version__ import __app_version__, __app_package__

# ~ Options ~~~~~

options = dict() # type: Dict[str, Dict[str, Any]]

options['external'] = {
    'keys': ['-e', '--external'],
    'params': {
        'default': False,
        'is_flag': True,
        'multiple': False,
        'show_default': True,
        'help': 'Enables external resources check',
    }
}

options['retry'] = {
    'keys': ['-r', '--retry'],
    'params': {
        'default': 0,
        'type': click.IntRange(0, 10),
        'is_flag': False,
        'multiple': False,
        'show_default': True,
        'help': 'try for error',
    }
}
options['threads'] = {
    'keys': ['-n', '--threads'],
    'params': {
        'default': 1,
        'type': click.IntRange(1, 10),
        'is_flag': False,
        'multiple': False,
        'show_default': True,
        'help': 'Specify number of parallel processes to use',
    }
}
options['domains'] = {
    'keys': ['-d', '--domains'],
    'params': {
        'multiple': True,
        'help': 'Domains to ignore',
    },
}
options['pathes'] = {
    'keys': ['-p', '--pathes'],
    'params': {
        'multiple': True,
        'help': 'Pathes to ignore',
    },
}

options['show'] = {
    'keys': ['-s', '--show'],
    'params': {
        'default': ['failed'],
        'multiple': True,
        'type': click.Choice([
            'failed',
            'ok',
            'ignored',
            'all',
            'none',
        ], case_sensitive=False),
        'show_default': True,
        'help': 'Category of URLs to show.',
    },
}

options['export'] = {
    'keys': ['--export'],
    'params': {
        'default': 'default',
        'hidden': True,
        'multiple': False,
        'type': click.Choice([
            'default',
        ], case_sensitive=False),
        'help': 'Export type',
    },
}

options['within_path'] = {
    'keys': ['--full-site-check'],
    'params': {
        'default': False,
        'is_flag': True,
        'help': 'Check links on domain not limiting',
    },
}

options['color'] = {
    'keys': ['--no-colors'],
    'params': {
        'default': False,
        'is_flag': True,
        'help': 'Color output of `default` export',
    },
}

# ~ Actual command decoration ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@click.command(__app_package__, cls=CLI, context_settings={"ignore_unknown_options": True})
@click.argument('url', nargs=1, required=True, metavar='<URL>')
@click.option(*options['external']['keys'], **options['external']['params'])
@click.option(*options['threads']['keys'], **options['threads']['params'])
@click.option(*options['retry']['keys'], **options['retry']['params'])
@click.option(*options['domains']['keys'], **options['domains']['params'])
@click.option(*options['pathes']['keys'], **options['pathes']['params'])
@click.option(*options['show']['keys'], **options['show']['params'])
@click.option(*options['export']['keys'], **options['export']['params'])
@click.option(*options['within_path']['keys'], **options['within_path']['params'])
@click.option(*options['color']['keys'], **options['color']['params'])
@click.version_option(
    message='%(prog)s: v%(version)s',
    prog_name=__app_package__,
    version=__app_version__,
)
@click.pass_context
def main(ctx: click.Context, url: str, **opts: Dict[str, Any]) -> None:
    """ checking links from web resource for dead/alive status. """

    try:
        settings = Settings(
            normilize_domain(url),
            **{
                'check_external_urls': opts['external'],
                'stay_within_path': not opts['full_site_check'],
                'threads': opts['threads'],
                'retry': opts['retry'],
                'ignore_domains': opts['domains'],
                'ignore_pathes': opts['pathes'],
            },
        )
        crawler = Crawler(settings)
        crawler.start()

        if str(opts['export']) == "default":
            Console(crawler, **opts).report()

    except DeadlinksExeption as e:
        ctx.fail(str(e))


def normilize_domain(url: str) -> str:
    """ 'guessing' scheme for urls without it. """

    u = Link(url)
    if not u.is_valid() and not u.scheme:
        return "http://{}".format(url)

    return url


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main()



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.exceptions
~~~~~~~~~~~~~~~~~~~~

Exceptions used for deadlinks package.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------


class DeadlinksExeption(BaseException):
    """ General Exception """


class DeadlinksRedicrectionURL(DeadlinksExeption):
    """ Redicrection URL """


class DeadlinksIgnoredURL(DeadlinksExeption):
    """ Error when we trying to index ignored URL """


class DeadlinksSettings(DeadlinksExeption):
    """ Error on Settings object """


class DeadlinksSettingsBase(DeadlinksSettings):
    """ Error on Settings object related to `BaseUrl` property """


class DeadlinksSettingsThreads(DeadlinksSettings):
    """ Error on Settings object related to `threads` property """


class DeadlinksSettingsRetry(DeadlinksSettings):
    """ Error on Settings object related to `retry` property """


class DeadlinksSettingsChange(DeadlinksSettings):
    """ Error on Settings object related to settings change  """


class DeadlinksSettingsDomains(DeadlinksSettings):
    """ Error on Settings object related to `IgnoredDomains` property """


class DeadlinksSettingsPathes(DeadlinksSettings):
    """ Error on Settings object related to `IgnoredPathes` property """


class DeadlinksSettingsPath(DeadlinksSettings):
    """ Error on Settings object related to `StayWithinPath` property """



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.main
~~~~~~~~~~~~~~~~~

main (cli interface)

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# package meta data
__app_version__ = "0.0.2"
__app_package__ = "deadlinks"
__app_license__ = "Apache License 2.0"
__app_website__ = "https://github.com/butuzov/deadlinks"
__description__ = "CLI/API for links liveness checking."
__author_name__ = "Oleg Butuzov"
__author_mail__ = "butuzov@made.ua"

# backwards compatibility
__version__ = __app_version__



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.__init__
~~~~~~~~~~~~~~~~~~

deadlinks module __init__ file

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from deadlinks.url import URL
from deadlinks.help import CLI
from deadlinks.link import Link
from deadlinks.index import Index
from deadlinks.crawler import Crawler
from deadlinks.request import request
from deadlinks.settings import Settings

from deadlinks.__version__ import __version__ as version

from deadlinks.exceptions import (
    DeadlinksIgnoredURL,
    DeadlinksSettingsThreads,
    DeadlinksSettingsBase,
    DeadlinksSettingsRetry,
    DeadlinksSettingsChange,
    DeadlinksSettingsDomains,
    DeadlinksSettingsPathes,
    DeadlinksSettingsPath,
)

__all__ = [
    'version',

    # App
    'URL',
    'CLI',
    'Link',
    'Index',
    'Crawler',
    'request',
    'Settings',

    # Exceptions
    'DeadlinksIgnoredURL',
    'DeadlinksSettingsThreads',
    'DeadlinksSettingsBase',
    'DeadlinksSettingsRetry',
    'DeadlinksSettingsChange',
    'DeadlinksSettingsDomains',
    'DeadlinksSettingsPathes',
    'DeadlinksSettingsPath',
]



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.url
~~~~~~~~~~~~~

URL representation with benefits

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (List, Optional) #pylint: disable-msg=W0611

from urllib.parse import (urlparse, urljoin)
from re import compile as _compile

from requests import RequestException

from deadlinks.request import request
from deadlinks.status import Status
from deadlinks.exceptions import (
    DeadlinksIgnoredURL,
    DeadlinksRedicrectionURL,
)

# -- Constants -----------------------------------------------------------------

__RE_LINKS__ = _compile(r'<a\s{1}([^>]+)>') # pylint: disable=W1401

# filters
CLEANER = lambda x: x.strip("\"'\n ") # removes quotes, spaces and new lines
ANCHORS = lambda x: x.split("#")[0] # removed part after anchor


class URL:
    """ URL abstraction representation. """

    def __init__(self, location: str) -> None:
        self._url = urlparse(location)
        self._status = Status.UNDEFINED # type: Status

        # some predefined states
        self._referrers = [] # type: List[str]
        self._text = None # type: Optional[str]
        self._links = [] # type: List[str]

        # internal error or mesage field, used to store ignore message
        # or error or status code.
        # TODO - rethink logick behined this value.
        self._message = "" # type: str

    # Basic properies of the URL Link
    @property
    def domain(self) -> str:
        """ Short netlocation prop. """
        return self._url.netloc

    @property
    def scheme(self) -> str:
        """ Short scheme prop. """
        return self._url.scheme

    @property
    def path(self) -> str:
        """ Short path prop. """
        return self._url.path

    @property
    def status(self) -> Status:
        """ Return one of 4 statuses of the URL ."""
        return self._status

    @status.setter
    def status(self, value: Status) -> None:
        """ Setter for status property. """
        if not isinstance(value, Status):
            raise TypeError("URL Status value can have only Status type ")
        self._status = value

    @property
    def message(self) -> str:
        return self._message

    @message.setter
    def message(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("message can be only string")
        self._message = value

    def is_valid(self) -> bool:
        """ Check if url looks "valid" """

        return (self.domain != "" and self.scheme in {"http", "https"})

    def add_referrer(self, url: str) -> None:
        """ Add a page that links (refferer) to self object. """
        if url in self._referrers:
            return
        self._referrers.append(url)

    def get_referrers(self) -> List[str]:
        """ Return URL refferers list. """
        return self._referrers

    def match_domains(self, domains: List[str]) -> bool:
        """ Match ignored pathes (argument pathes) to `url.netloc`. """
        for domain in domains:
            if domain in self._url.netloc:
                return True
        return False

    def match_pathes(self, pathes: List[str]) -> bool:
        """ Match ignored pathes (argument pathes) to `url.path`. """
        for path in pathes:
            if path in self._url.path:
                return True
        return False

    def exists(self, is_external: bool = False, retries: int = 0) -> bool:
        """ Return "found" (or "not found") status of the page as bool. """

        if self.status in (Status.FOUND, Status.NOT_FOUND):
            return True if Status.FOUND else False

        if self.status == Status.IGNORED:
            error = "This URL <{}> ignored"
            raise DeadlinksIgnoredURL(error.format(self.url()))

        try:
            response = request(self.url(), is_external, retries)
        except RequestException as exception:
            self.message = str(exception)
            return False

        # Group of 2XX responses. In general we think its OK to mark URL as
        # reachable and exists
        if response.status_code // 100 == 2:
            self._text = response.text
            return True

        # redirections catching.
        if response.status_code // 100 == 3:
            raise DeadlinksRedicrectionURL(response.headers['location'])

        self.message = str(response.status_code)
        return False

    def url(self) -> str:
        """ Return url based on abstraction, minus ending slash. """
        return self._url.geturl()

    def __str__(self) -> str:
        """ Converts URL object to string (actual URL). """
        return self.url()

    def __repr__(self) -> str:
        """ Object stringer representation. """

        return "{}<{}>".format(self.__class__.__name__, self.url())

    def consume_links(self) -> None:
        """ Parse response text into list of links. """

        if not self._text:
            return

        links = []
        for attr in __RE_LINKS__.findall(self._text):
            pos = attr.find("href=")
            if pos == -1:
                "href not found"
                continue

            href = attr[pos + 5:].strip()

            if not href:
                continue

            link = "" # type: str

            quoted = href[0] in {'"', "'"}
            if quoted:
                end_pos = href[1:].find(href[0])
                if end_pos == -1:
                    "unquoted link"
                    continue
                link = href[1:end_pos + 1]
            else:
                end_pos = href[0:].find(" ")
                link = href if end_pos == -1 else href[:end_pos + 1]

            if not link:
                "empty link"
                continue

            links.append(link)

        self._links = list(map(ANCHORS, map(CLEANER, links)))

    @property
    def links(self) -> List[str]:
        """ Return links found at the page. """

        if not self._text or self._links:
            return self._links

        self.consume_links()

        return self._links

    def link(self, href: str) -> str:
        """
        Construct a full (“absolute”) URL by combining a
        “URL” object as base with another URL (url).

        note: we not going to use self.url() as it removes ending slash.
        """

        return urljoin(self._url.geturl(), href)


if __name__ == "__main__": # pragma: no cover
    urls = set()
    urls.add(URL("http://google.com"))
    urls.add(URL("http://google.com"))
    urls.add(URL("https://google.com"))

    print(len(urls), urls)



"""
deadlinks
~~~~~~~~~
deadlinks checker for your static website. It's better keep house clean, right?
"""

from typing import (Dict, List) #pylint: disable-msg=W0611

from pathlib import Path
from re import compile as _compile
from collections import defaultdict
import sys, os
from setuptools import find_packages, setup


def read_data() -> Dict[str, str]:
    init = Path(__file__).parent / "deadlinks" / "__version__.py"

    if not Path(init).is_file():
        raise RuntimeError("Can not find source for deadlinks/__version__.py")

    DUNDER_REGEXP = _compile(r'(__(.*?)__ = "(.*?)")\n')

    values = dict() # type: Dict[str, str]

    with open(str(init)) as fh:
        content = "".join(fh.readlines())
        for match in DUNDER_REGEXP.findall(content):
            values[match[1]] = match[2]

    return values


def require(section: str = "install") -> List[str]:
    """requirements txt parser"""

    require_txt = Path(__file__).parent / "requirements.txt"
    if not Path(require_txt).is_file():
        return []

    requires = defaultdict(list) # type: Dict[str, List[str]]

    with open(str(require_txt), "rb") as fh:
        key = "" # type: str
        for line in fh.read().decode("utf-8").split("\n"):

            if not line.strip():
                " empty line "
                continue

            if line[0] == "#":
                " section key "
                key = line[2:]
                continue

            # actual package
            requires[key].append(line.strip())

    return requires[section]


def readme() -> str:
    """ different version of readme changed for pypi """
    readme = Path(__file__).parent / "README.md"

    contents = ""

    if not Path(readme).is_file():
        return contents

    with open("README.md", encoding="utf8") as f:
        contents = f.read()

    # cutout first header

    contents = contents.replace("# deadlinks", "", 1).lstrip()

    return contents


# ------------------------------------------------------------------------------
data = read_data()

# - Overriding version for test deployment

if os.environ.get('DEADLINKS_VERSION', None):
    data['app_version'] += "."
    data['app_version'] += str(os.environ.get('DEADLINKS_VERSION'))

# - Local testing
TESTS = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
TESTS_RUNNER = ['pytest-runner'] if TESTS else []

# Setup
setup(
    name=data['app_package'],
    version=data['app_version'],
    description=data['description'],
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords=["documentation", "website", "spider", "crawler", "link-checker"],
    author=data['author_name'],
    author_email=data['author_mail'],
    packages=find_packages(exclude=["tests*"]),
    install_requires=require("install"),
    tests_require=require("tests"),
    setup_requires=(require("install") + TESTS_RUNNER),
    test_suite="pytest",
    extras_require={
        'test': require("tests"),
        'all': require("install") + require("tests") + require("linters"),
        'lint': require("linters"),
    },
    entry_points='''
        [console_scripts]
        deadlinks=deadlinks.__main__:main
    ''',
    zip_safe=False,
    python_requires='>=3.5',
    url=data['app_website'],
    license=data['app_license'],
    platforms=['MacOS', 'Posix', 'Unix'],
    classifiers=[
        # Env
        "Environment :: Console",

        # Status
        "Development Status :: 2 - Pre-Alpha",

        # Audience
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",

        # Topic
        "Topic :: Utilities",
        "Topic :: Documentation",

        # Audience and Topic
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",

        # Python version
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",

        # License
        "License :: OSI Approved :: Apache Software License",

        # Operation System
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Unix",

        # Language
        "Natural Language :: English",
    ],
)



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.request
~~~~~~~~~~~~~~~~~

`requests` wrapper for getting requests.Response

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from requests import Session, Response
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from deadlinks.__version__ import __app_version__ as version
from deadlinks.__version__ import __app_website__ as website
from deadlinks.__version__ import __app_package__ as package


def request(url: str, is_external: bool = False, retries_attempts: int = 1) -> Response:
    r"""Request a web resource and return Response

    Perform GET  - for the local resource
            HEAD - for the remote resource
    Return Response of the request.
    """

    _headers = {
        'User-agent': "{}/v{} ( {} )".format(package, version, website),
    }
    _settings = {
        'allow_redirects': False,
    }

    _retry = Retry(
        total=retries_attempts,
        backoff_factor=1,
        status_forcelist=[502, 503, 504],
    )

    session = Session()
    session.mount(url, HTTPAdapter(max_retries=_retry))

    method_to_call = (session.head if is_external else session.get)
    return method_to_call(url, headers=_headers, **_settings)
